#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentTypeError, \
                     RawDescriptionHelpFormatter, BooleanOptionalAction
import sys
import re
from functools import reduce
import h5py
import numpy as np
import matplotlib.pyplot as plt
import scipy.constants
from ase.data import chemical_symbols
from vasp_suite.structure import Structure

import xyz_py as xyzp
from molcas_suite.cli import ParseExtra
import hpc_suite as hpc
from hpc_suite.action import ParseKwargs, OrderedSelectionAction
from hpc_suite import make_parse_dict, SecondaryHelp, parse_index
import angmom_suite as angmom
from angmom_suite.crystal import calc_oef, calc_total_strength
from angmom_suite.basis import Level
from angmom_suite.multi_electron import Ion
from gaussian_suite.extractor import make_extractor as make_gaussian_extractor
import env_suite as env

from .derivative import print_tau_style, read_tau_style
from .lvc import LVC, generate_lvc_input, ANG2BOHR, \
    make_lvc_evaluator_eq, make_lvc_evaluator_xyz, make_lvc_evaluator_hta, \
    LVCModelHamiltonian, LVCAdiabaticEnergies, LVCDiabaticHamiltonian
from .vibrations import Harmonic, HARTREE2J, BOHR2M, E0_TO_C, construct_atomic_basis
from .polarisation_derivatives import parse_BEC, normalised_sphere_vectors
from .experiments import make_exp_evaluator

plt.rcParams['font.family'] = "Arial"

HARTREE2INVCM = 219474.63
EPSILON0 = scipy.constants.epsilon_0
BOHR_RADIUS = scipy.constants.physical_constants["Bohr radius"][0]

BOHR2_TO_COULOMBMETERPERVOLT = 1.0 / (4 * np.pi * EPSILON0 * BOHR_RADIUS**2)


# Action for secondary function help message
class FunctionHelp(SecondaryHelp):
    def __call__(self, parser, namespace, values, option_string=None):

        if namespace.function == 'proj':
            angmom.read_args(['proj', '--help'])
        else:
            raise ValueError("Supply valid option to function argument.")


def str_int(x):
    try:
        return int(x)
    except ValueError:
        return str(x)


def lvc_func(args):

    if args.add is not None:
        def join(a, b):
            return a.join(b, xyzp.load_xyz(args.reference_xyz)[1] * ANG2BOHR)
        lvc = reduce(join, (LVC.from_file(f_lvc) for f_lvc in args.add))

    elif args.efld is not None:
        charge_list = [np.loadtxt(f_chrg) for f_chrg in args.charge]
        xyz_list = [xyzp.load_xyz(f_xyz)[1] * ANG2BOHR for f_xyz in args.xyz]
        lvc = LVC.from_molcas_frozen_density(
            args.rassi, args.efld, charge_list, xyz_list)

    elif args.grad is not None:
        lvc = LVC.from_molcas(args.rassi, args.grad)

    else:
        raise ValueError("LVC gradient information supplied!")

    if args.active_atoms is not None:
        lvc = lvc.subset_atoms(active_atom_idc=args.active_atoms)

    elif args.ref_coordinates is not None:
        _, coords = xyzp.load_xyz(args.ref_coordinates)
        lvc = lvc.subset_atoms(ref_coords=coords * ANG2BOHR)

    lvc.to_file(args.lvc_data, verbose=args.verbose)

    return


def eval_func(args, unknown_args):

    try:
        selected = args._selection
    except AttributeError:
        sys.exit("No quantity selected for evaluation!")

    # Resolve coordinates
    if args.vibration_info is not None:
        ho = Harmonic.from_file(args.vibration_info)
        if args.hta:
            disp = ho.displacements
            std_dev = ho.freqs
            coords = None
        else:
            coords = ho.mass_freq_weighted_coordinates * ANG2BOHR
    else:
        coords = None

    if args.geom:
        evaluator = make_lvc_evaluator_xyz
        store_args = hpc.read_args(['store'] + unknown_args + ['-i'] + args.geom + ['--group'] + args.geom)
    elif args.hta:
        evaluator = make_lvc_evaluator_hta
        store_args = hpc.read_args(['store'] + unknown_args)
    else:
        evaluator = make_lvc_evaluator_eq
        store_args = hpc.read_args(['store'] + unknown_args)

    if args.hta:
        hpc.store_func(
            store_args,
            evaluator,
            selected,
            lvc=LVC.from_file(args.lvc_data),
            order=args.order,
            truncate=args.truncate,
            align=args.align,
            n_geom=args.hta,
            temp=args.temp,
            sep=args.individual_evals,
            disp=disp,
            std_dev=std_dev,
            coords=coords
        )
    else:
        hpc.store_func(
            store_args,
            evaluator,
            selected,
            lvc=LVC.from_file(args.lvc_data),
            order=args.order,
            coords=coords,
            truncate=args.truncate,
            align=args.align
        )


def vib_func(args):


    if args.gaussian_log:

        if args.mass:
            raise ValueError("The --mass argument is not available with freq "
                             "data read from Gaussian text output use *.fchk!")

        ho = Harmonic.from_gaussian_log(args.gaussian_log)

    elif args.gaussian_fchk:
        ho = Harmonic.from_gaussian_fchk(
            args.gaussian_fchk,
            ext_masses=args.mass,
            rot=args.rot,
            trans=args.trans
        )

    elif args.orca_hess:
        ho = Harmonic.from_orca_hess(
            args.orca_hess,
            ext_masses=args.mass,
            rot=args.rot,
            trans=args.trans
        )

    elif args.vasp_outcar:
        ho = Harmonic.from_vasp_outcar(
                args.vasp_outcar,
                args.poscar,
                trans=args.trans
        )

    elif args.force_sets:

        ho = Harmonic.from_vasp_phonopy(
            args.poscar,
            args.force_sets,
            ext_masses=args.mass,
            trans=args.trans,
            cluster_expansion=args.cluster_expansion,
            cluster_cutoff=args.cluster_cutoff,
            central_idc=args.central_index,
            force_expansion=args.force_expansion,
            q_mesh=args.q_mesh
        )

    elif args.phonopy_hdf5:

        if args.mass:
            raise ValueError("The --mass argument is not available with phonon"
                             " data read from phonopy hdf5 database, use "
                             "FORCE_SETS instead!")

        ho = Harmonic.from_phonopy_hdf5(
            args.poscar,
            args.phonopy_hdf5,
            trans=args.trans,
            cluster_expansion=args.cluster_expansion,
            cluster_cutoff=args.cluster_cutoff,
            central_idc=args.central_index
        )

    if args.imaginary=='remove':
        ho = ho.remove_imaginary(verbose=True)
    elif args.imaginary=='absolute':
        ho = ho.abs_imaginary(verbose=True)

    if args.active_atoms is not None:
        ho = ho.subset_atoms(active_atom_idc=args.active_atoms)

    elif args.ref_coordinates is not None:
        _, coords = xyzp.load_xyz(args.ref_coordinates)
        ho = ho.subset_atoms(ref_coords=coords)

    ho.to_file(args.vibration_info, store_vecs=args.store_vecs)

    if args.save_pyvibms is not None:
        ho.to_pyvibms(args.save_pyvibms)

    return


def prepare_func(args):

    # mode_energies.dat
    ho = Harmonic.from_file(args.vibration_info)

    np.savetxt('mode_energies.dat', ho.freqs, fmt='%16.8f',
               header="Frequency (cm-1)", comments='')

    name = args.program.split('_')[0]

    if name == 'firms' and args.optical_info is None:
        raise ValueError("--optical_info must be included in ['firms', 'firms_direct']")

    if args.program.split('_')[-1] == 'direct':
        lvc = LVC.from_file(args.lvc_data)

        ener = LVCAdiabaticEnergies(lvc, 0, soc=True,
                                    truncate=args.truncate)[()]

        nstates = args.truncate or ener.size

        couplings = LVCDiabaticHamiltonian(lvc, 1,
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            soc=True, truncate=nstates)

        with h5py.File(f'{name}.hdf5', 'w') as h, open('coupling_strength.dat', 'w') as f:

            h.create_dataset('energies', data=(ener - ener[0]) * HARTREE2INVCM)
            h.create_dataset('mode_energies', data=ho.freqs)

            grp = h.create_group("couplings")

            for modes, lin in iter(couplings):

                cplg = lin * HARTREE2INVCM

                if args.trace_less:
                    cplg -= np.identity(nstates) * np.trace(cplg) / nstates

                grp.create_dataset('_'.join(map(str, modes)), data=cplg)
                strength = np.sum((cplg * cplg.conj()).real)
                f.write(f'{strength}\n')

            angm_grp = h.create_group('angmom')
            spin_grp = h.create_group('spin')

            for idx, comp in enumerate(["x", "y", "z"]):
                idc = np.ix_(range(nstates), range(nstates))
                angm_grp.create_dataset(comp, data=lvc.sos_angm[idx][idc])
                spin_grp.create_dataset(comp, data=lvc.sos_spin[idx][idc])

    if args.program in ('tau', 'firms'):
        # CFP args
        proj_kwargs = {
            'model_space': args.ground_level,
            'quax': args.quax,
            'terms': {"cf": [('J',)]},
            'k_max': args.k_max,
            'theta': True,
            'ion': args.ion
        }

        # EQ_CFPs.dat
        cfp_eq = LVCModelHamiltonian(LVC.from_file(args.lvc_data), 0,
                                     **proj_kwargs)[("cf", "J")]

        cfp_data = np.column_stack(
            (np.ones(27), list(cfp_eq.keys()), list(cfp_eq.values())))

        np.savetxt('EQ_CFPs.dat', cfp_data,
                   fmt=['%2d', '%2d', '% 2d', '%16.8f'])

        cfp_dq = LVCModelHamiltonian(
            LVC.from_file(args.lvc_data), 1,
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            **proj_kwargs
        )

        print_tau_style(filter(lambda pars: pars[0][1] == 'cf', iter(cfp_dq)),
                        ho.freqs, 'CFP_polynomials.dat')

    elif name == 'firms':
        if args.vasp_born:
            bec_tensors = parse_BEC(args.poscar, args.vasp_born)
            derivatives = ho.calculate_dmudQ(bec_tensors) # [C amu^{-1/2}]
            red_masses = ho.mode_red_masses
            derivatives *= np.sqrt(red_masses).reshape((-1, 1)) # [C]
        elif args.gaussian_log:
            derivatives = make_gaussian_extractor(args.gaussian_log, ("freq", "dipole_derivatives"))[()]
            derivatives *= 1/np.sqrt(42.2561) * 3.33564e-30 * 1e10 * np.sqrt(ho.red_masses)
        elif args.gaussian_fchk:
            derivatives = make_gaussian_extractor(args.gaussian_fchk, ("fchk", "dipole_derivatives"))[()]
            derivatives = derivatives.reshape(-1, 3, 3) * E0_TO_C  # [C amu^{-1/2}]
            if args.active_atoms:
                derivatives = derivatives[args.active_atoms, :, :]
            derivatives = np.einsum('jab, kja -> kb', derivatives.reshape(-1, 3, 3), ho.displacements)
            red_masses = np.loadtxt(args.red_masses)
            derivatives *= np.sqrt(red_masses).reshape((-1,1)) # [C]

        # read in the optical properties
        with h5py.File(args.optical_info, 'r') as optics_hdf5:
            ESO = np.array(optics_hdf5['ESO'][()], dtype=np.complex128)
            EDTM = np.array(optics_hdf5['EDTM'][()], dtype=np.complex128)
            MDTM = np.array(optics_hdf5['MDTM'][()], dtype=np.complex128)

        nstates = args.truncate or ESO.size

        ESO = ESO[:nstates]
        EDTM = EDTM[:, :nstates, :nstates]
        MDTM = MDTM[:, :nstates, :nstates]

        with h5py.File('firms.hdf5', 'a') as h:
            h.create_dataset(name='dipole_derivatives', data=derivatives)
            h.create_dataset(name='red_masses', data=red_masses)
            h.create_dataset(name='ESO', data=ESO)
            h.create_dataset(name='EDTM', data=EDTM)
            h.create_dataset(name='MDTM', data=MDTM)

    return


def generate_lvc_input_func(args):

    # todo: replace dict by global definition
    roots_dict = {'dy': [18]}

    if args.num_roots:
        num_roots = map(int, args.num_roots)
    elif args.ion:
        num_roots = roots_dict[args.ion.lower()]
    else:
        sys.exit("Invalid specification of the number of roots.")

    for jobiph_idx, num_root in enumerate(num_roots):

        if num_root == 0:
            continue

        generate_lvc_input(
            args.old_path,
            args.old_project,
            num_root,
            jobiph_idx + 1,
            jobiph_path=args.jobiph_paths[jobiph_idx] if args.jobiph_paths is not None else None,
            mclr_extra=args.mclr_extra,
            alaska_extra=args.alaska_extra,
            two_step=args.two_step,
            dry=args.dry
        )

    return


def strength_func(args):
    """
    Wrapper for spin-phonon coupling strength CLI call
    """

    # Load CFP polynomials from file
    dQ, freqs = read_tau_style(args.cfp_file)

    # TODO delete
    args.n = 13
    args.J = 3.5
    args.L = 3
    args.S = 0.5

    if not args.nooef:
        # Get OEF values
        OEFs = calc_oef(args.n, args.J, args.L, args.S)
        print(np.shape(OEFs))

        # Add back in OEFs to polynomial values
        dQ *= OEFs

    # Calculate strength values of each mode
    S = np.array([calc_total_strength(mode) for mode in dQ])

    # Save strength to file
    np.savetxt("strengths.dat", S)
    print("Strengths saved to strengths.dat")

    # Read in symmetry labels if given
    if args.irreps:
        irreps = list(np.loadtxt(args.irreps, dtype=str))
    else:
        irreps = ['A']*len(freqs)

    unique_irreps, irrep_ints = np.unique(irreps, return_inverse=True)

    if args.plot:

        _, ax = plt.subplots(num='Spin-phonon coupling strength')

        for unique_int in np.unique(irrep_ints):
            ax.stem(
                freqs[np.nonzero(irrep_ints == unique_int)[0]],
                S[np.nonzero(irrep_ints == unique_int)[0]],
                basefmt=' ',
                linefmt='C{:d}-'.format(unique_int),
                markerfmt='C{:d}o'.format(unique_int)
            )
        if args.irreps:
            ax.legend(unique_irreps, frameon=False)

        ax.set_ylabel(r'$S$ (cm$^{-1}$)', fontname="Arial")
        ax.set_xlabel('Mode energy (cm$^{-1}$)', fontname="Arial")

        ax.set_ylim([0., np.max(S)*1.05])

        plt.savefig("strengths.svg")
        plt.savefig("strengths.png")
        print("Strength plots saved to strengths.svg and strengths.png")
        plt.show()

    return


def exp_func(args, unknown_args):
    """
    Wrapper for experimental data simulation CLI call
    """

    try:
        selected = args._selection
    except AttributeError:
        sys.exit("No experiment selected!")

    store_args = hpc.read_args(['store'] + unknown_args + ['-i'] + args.vibration_info)

    hpc.store_func(
        store_args,
        make_exp_evaluator,
        selected
    )

    return


def electric_field_func(args):
    """
    Model distortions due to an applied electric field from a single point
    frequency calculation in Gaussian16 or VASP

    Equation (Gas phase):
        -2 U^T H U U^T r = U^T ( 1/1! nabla_r mu * E + 1/2! nabla_r alpha * E^2 + ... )

    Equation (Solid state):
        -2 U^T H U U^T r = U^T (nabla_r mu * E)
    """

    if args.gaussian_fchk is not None:
        hess = make_gaussian_extractor(args.gaussian_fchk, ('fchk', 'hessian'))[()]
        coords = make_gaussian_extractor(args.gaussian_fchk, ('fchk', 'coordinates'))[()]
        grad_mu = make_gaussian_extractor(args.gaussian_fchk, ('fchk', 'dipole_derivatives'))[()]
        numbers = make_gaussian_extractor(args.gaussian_fchk, ('fchk', 'atomic_number'))[()]
        symbols = list(map(lambda x: chemical_symbols[x], numbers))

        # unit conversion
        hess *= HARTREE2J / BOHR2M**2  # J/m^2 = N/m
        grad_mu *= E0_TO_C  # C
        coords *= BOHR2M * 1e10  # Angstrom

        U = construct_atomic_basis(int(grad_mu.shape[0]/3), gaussian_fchk=True,
                                   coords=coords)  # units in angstrom
        hess = U.T @ hess @ U

        if args.field_vector is not None:
            E = np.array(args.field_vector)
            grad_mu_E = grad_mu @ E
            r = np.linalg.solve(-1 * hess, U.T @ grad_mu_E)
            r = (U @ r) * 1e10

            new_coords = coords + r.reshape(-1, 3)
            xyzp.save_xyz(args.output, symbols, new_coords)

        elif args.range is not None:
            args.range = args.range[0].split(' ')
            args.range = list(map(float, args.range))
            start, stop, step, dir = args.range
            field_range = np.linspace(start, stop, int(step))
            field_vector = np.eye(3)[int(dir)]
            for idx, field_strength in enumerate(field_range):
                E = field_strength * field_vector
                grad_mu_E = grad_mu @ E
                r = np.linalg.solve(-1 * hess, U.T @ grad_mu_E)
                r = (U @ r) * 1e10

                output = args.output.split('.')
                output = output[0] + '_{}'.format(idx) + '.' + output[1]
                xyzp.save_xyz(output, symbols, coords + r.reshape(-1, 3))

        elif args.field_sampling is not None:
            num_points = int(args.field_sampling[0])
            field_strength = args.field_sampling[1]
            field_vectors = field_strength * np.unique(normalised_sphere_vectors(num_points), axis=0)
            print("Number of sampled electric field {} of strength {} V/m".format(num_points, field_strength))
            with open('sample_fields.dat', 'w') as f:
                for idx, E in enumerate(field_vectors):
                    grad_mu_E = grad_mu @ E
                    r = np.linalg.solve(-2 * hess, U.T @ grad_mu_E)
                    r = (U @ r) * 1e10

                    output = args.output.split('.')
                    output = output[0] + '_{}'.format(idx) + '.' + output[1]
                    xyzp.save_xyz(output, symbols, coords + r.reshape(-1, 3))
                    f.write('{} {} {}\n'.format(*E))
            print('Sampled fields saved to sample_fields.dat')

    elif args.vasp_born is not None:
        struct = Structure.from_poscar(args.poscar)
        coords = struct.cart_coords
        symbols = struct.atom_list
        grad_mu = parse_BEC(args.poscar, args.vasp_born)

        # reshape of Z^*_{i,ab} -> dmu_{a}/dr_{i,b}
        # [i, a, b] -> [i, b, a] -> [ib, a] == [3N x 3]
        grad_mu = np.array(list(map(lambda x: x.T, grad_mu)))
        grad_mu = grad_mu.reshape(-1, 3)

        vibration = Harmonic.from_file(args.vibrational_info)
        hess = vibration.vasp_hessian
        U = construct_atomic_basis(int(hess.shape[0]/3), vasp_born=True)
        hess = U.T @ hess @ U
        # unit conversion
        grad_mu *= E0_TO_C  # C
        if not args.range:
            grad_mu_E = grad_mu @ np.array(args.field_vector)
            r = np.linalg.solve(hess, -2 * U.T @ grad_mu_E)
            r = U @ r
            r *= 1e10  # Å
            new_coords = coords + r.reshape(-1, 3)
            frac_coords = new_coords @ np.linalg.inv(struct.lattice_vectors)
            frac_coords %= 1
            struct.coords = frac_coords
            struct.write_poscar(args.output)
        else:
            start, stop, step, dir = args.range
            field_range = np.linspace(start, stop, int(step))
            field_vector = np.eye(3)[int(dir)]
            for field_strength in field_range:
                efield = field_strength * field_vector
                grad_mu_E = grad_mu @ efield

                r = np.linalg.solve(-2 * hess, U.T @ grad_mu_E)
                r = U @ r
                r *= 1e10  # Å
                output = args.output.split('.')
                output = output[0] + '_{:1.3e}'.format(field_strength) + '.' + output[1]
                frac_coords = (coords + r.reshape(-1, 3)) @ np.linalg.inv(struct.lattice_vectors)
                frac_coords %= 1
                struct.coords = frac_coords
                struct.write_poscar(output)

    else:
        raise NotImplementedError


def read_args(arg_list=None):
    description = '''
    A package for performing Spin-Phonon coupling calculations.
    '''

    parser = ArgumentParser(
            description=description,
            formatter_class=RawDescriptionHelpFormatter
            )

    subparsers = parser.add_subparsers(dest='prog')

    evaluate = subparsers.add_parser('eval')
    evaluate.set_defaults(func=eval_func)

    evaluate.add_argument(
        '-H',
        '--Help',
        action=FunctionHelp,
        help='show help message for additional arguments and exit'
    )

    evaluate.add_argument(
        '-L',
        '--lvc_data',
        type=str,
        help='HDF5 database containing the LVC parameters.'
    )

    evaluate.add_argument(
        '-V',
        '--vibration_info',
        type=str,
        help=('HDF5 database containing information about the vibrations.'
              'Derivatives are evaluated in the basis of dimension-less'
              'mass-frequency weighted normal mode coordinates.')
    )

    evaluate.add_argument(
        '--geom',
        nargs='+',
        help='*.xyz coordinates at which properties will be evaluated.'
    )

    evaluate.add_argument(
        '--order',
        type=int,
        default=0,
        help='Order of derivative.'
    )

    evaluate.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    evaluate.add_argument(
        '--align',
        action=BooleanOptionalAction,
        default=True,
        help='Align LVC model reference to input geometry by minimising RMSD.'
    )

    evaluate.add_argument(
        '--adiabatic_energies',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Energies of the MCH or SOC eigenstates.'
    )

    evaluate.add_argument(
        '--diabatic_hamiltonian',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Diabatic potential energy matrix between MCH or SOC states.'
    )

    evaluate.add_argument(
        '--proj',
        nargs='+',
        action=OrderedSelectionAction,
        help='Model Hamiltonian projection'
    )

    evaluate.add_argument(
        '--sus',
        nargs='+',
        action=OrderedSelectionAction,
        help='Magnetic susceptibility'
    )

    evaluate.add_argument(
        '--hta',
        type=int,
        metavar="n_geoms",
        help='Activate HTA averaging and specify number of geometries to generate.'
    )

    evaluate.add_argument(
        '--temp',
        type=float,
        metavar="temp",
        default=300.0,
        help='Temperature used for HTA averaging (Kelvin). Default 300 K.'
    )

    evaluate.add_argument(
        '--individual_evals',
        action=BooleanOptionalAction,
        default=False,
        help='Print properties from each HTA geometry as well as the average.'
    )

    inp = subparsers.add_parser('generate_input')
    inp.set_defaults(func=generate_lvc_input_func)

    inp.add_argument(
        'old_project',
        type=str,
        help='Project name of preceding Molcas calculation.'
    )

    inp.add_argument(
        '--old_path',
        type=str,
        default='../',
        help='Path to WorkDir of preceding Molcas calculation.'
    )

    roots = inp.add_mutually_exclusive_group(required=True)

    roots.add_argument(
        '--num_roots',
        nargs='+',
        help='Number of states per JOBIPH.'
    )

    roots.add_argument(
        '--ion',
        type=str,
        help='Label of the metal center, e.g. Dy.'
    )

    inp.add_argument(
        '--jobiph',
        nargs='+',
        help='Indices of Molcas JOBIPH wavefunction files *_IPH.'
    )

    inp.add_argument(
        '--jobiph_paths',
        nargs='+',
        help='Paths of Molcas JOBIPH wavefunction files *_IPH.'
    )

    inp.add_argument(
        '--mclr_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Manually run mclr with custom options, e.g. thre=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--alaska_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Run alaska with custom options, e.g. cuto=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--two_step',
        action=BooleanOptionalAction,
        default=True,
        help='Utilize two-step procedure for MCLR runs'
    )

    inp.add_argument(
        '--dry',
        default=False,
        action='store_true',
        help='Dry-run which prints files to be created'
    )

    lvc = subparsers.add_parser('lvc')
    lvc.set_defaults(func=lvc_func)

    lvc.add_argument(
        '-L', '--lvc_data',
        type=str,
        help='HDF5 database output containing the LVC data.'
    )

    grad = lvc.add_mutually_exclusive_group()

    grad.add_argument(
        '--grad',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing gradients and NACs.'
    )

    grad.add_argument(
        '--efld',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing electric field values.'
    )

    grad.add_argument(
        '--add',
        type=str,
        nargs='+',
        help='LVC HDF5 data bases to be added.'
    )

    frozen = lvc.add_argument_group('Frozen density gradient arguments')

    frozen.add_argument(
        '--charge',
        type=str,
        nargs='+',
        help='Text file(s) containing point charge values.'
    )

    frozen.add_argument(
        '--xyz',
        type=str,
        nargs='+',
        help='xyz file(s) containing point charge coordinates.'
    )

    add = lvc.add_argument_group('LVC addition arguments')

    add.add_argument(
        '--reference_xyz',
        type=str,
        help='xyz file containing the reference coordinates.'
    )

    lvc.add_argument(
        '--rassi',
        type=str,
        help=('Molcas *.rassi.h5 output file containing AMFI integrals, '
              'SF_ANGMOM operators and the spin multiplicities.')
    )

    subset = lvc.add_mutually_exclusive_group()

    subset.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_index,
        help=(
            'atomic indices active during spin-phonon coupling. effectively '
            'effectively setting the coupling of all other atoms to zero. '
            'useful to suppress coupling of specific atoms.'
        )
    )

    subset.add_argument(
        '--ref_coordinates',
        help='xyz coordinates containing the atomic positions of active atoms.'
    )

    lvc.add_argument(
        '--verbose',
        action='store_true',
        help='Plot gradient norm and invariance measures.'
    )

    vibrate = subparsers.add_parser('vib', parents=[env.cluster])
    vibrate.set_defaults(func=vib_func)

    vibrate.add_argument(
        '-V', '--vibration_info',
        type=str,
        help='HDF5 database containing information about the vibrations.'
    )

    vib_calc_excl = vibrate.add_mutually_exclusive_group(required=True)

    vib_calc_excl.add_argument(
        '--gaussian_log',
        type=str,
        help='Text output of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--gaussian_fchk',
        type=str,
        help='Formatted checkpoint file of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--orca_hess',
        type=str,
        help='Orca *.hess file',
    )

    vib_calc_excl.add_argument(
        '--force_sets',
        type=str,
        help='FORCE_SETS file from VASP-phonopy pre-process.'
    )

    vib_calc_excl.add_argument(
        '--phonopy_hdf5',
        type=str,
        help='Phonon database from phono3py calculation.'
    )

    vib_calc_excl.add_argument(
        '--vasp_outcar',
        type=str,
        help='OUTCAR file from IBRION=5 calculation.'
    )

    vibrate.add_argument(
        '--poscar',
        type=str,
        help='Unit cell POSCAR from VASP-phonopy pre-process.'
    )

    vibrate.add_argument(
        '--force_expansion',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        default=(1, 1, 1),
        type=int,
        help='Supercell expansion used in phonon calculation'
    )

    qpoints = vibrate.add_mutually_exclusive_group()

    qpoints.add_argument(
        '--q_mesh',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        type=int,
        help='Mesh in q-space for phonon evaluation.'
    )

    vibrate.add_argument(
        '--mass',
        type=make_parse_dict(str_int, float),
        default={},
        nargs='+',
        action=ParseKwargs,
        help='Modify atomic masses for isotopic substitution.',
        metavar='atom_index=mass or element_symbol=mass'
    )

    vibrate.add_argument(
        '--trans',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body translations.'
    )

    vibrate.add_argument(
        '--rot',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body rotations.'
    )

    vibrate.add_argument(
        '--imaginary',
        default='remove',
        choices=['remove', 'keep', 'absolute'],
        help=('Remove modes with imaginary frequencies (default), '
             'keep without altering, or set to their absolute value.')
    )


    subset = vibrate.add_mutually_exclusive_group()

    subset.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_index,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively'
            ' subsets the displacement vectors. Useful if coupling is '
            'evaluated with a subset of the atoms present in the vibrational'
            ' calculation.'
        )
    )

    subset.add_argument(
        '--ref_coordinates',
        help='xyz coordinates containing the atomic positions of active atoms.'
    )

    vibrate.add_argument(
        '--save_pyvibms',
        type=str,
        help='optional pyvibms output for visualisation in PyMol.'
    )

    vibrate.add_argument(
        '--store_vecs',
        action=BooleanOptionalAction,
        default=True,
        help='Flag to disable expensive storage of normal mode displacements.'
    )

    prepare = subparsers.add_parser('prep')
    prepare.set_defaults(func=prepare_func)

    prepare.add_argument(
        'program',
        choices=['tau', 'tau_direct', 'firms', 'firms_direct'],
        help='Program for which to prepare inputs.'
    )

    dip = prepare.add_mutually_exclusive_group()

    dip.add_argument(
            '--gaussian_log',
            type=str,
            help='Gaussian log file',
    )

    dip.add_argument(
            '--gaussian_fchk',
            type=str,
            help='Gaussian log file',
    )

    dip.add_argument(
            '--vasp_born',
            type=str,
            help='VASP BORN file',
    )

    prepare.add_argument(
            '--poscar',
            type=str,
            help='VASP POSCAR file, required for VASP BORN file',
    )

    data = prepare.add_argument_group('database files')

    data.add_argument(
        '-L', '--lvc_data',
        help='HDF5 database containing the LVC parameters.')

    data.add_argument(
        '-V', '--vibration_info',
        help='HDF5 database containing information about the vibrations.')

    data.add_argument(
        '-O', '--optical_info',
        help='HDF5 database containing information about the optical properties.',
        required=False,
        default=None)

    prepare.add_argument(
        '--ground_level',
        type=Level.parse,
        help='Symbol of the model space.'
    )

    prepare.add_argument(
        '--ion',
        type=Ion.parse,
        help='Central ion.'
    )

    prepare.add_argument(
        '--k_max',
        type=int,
        default=6,
        help='Maximum Stevens operator rank.'
    )

    prepare.add_argument(
        '--quax',
        action=angmom.QuaxAction,
        help='Quantisation axes.'
    )

    prepare.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    prepare.add_argument(
        '--trace_less',
        action=BooleanOptionalAction,
        default=True,
        help='Substract out trace from bare coupling elements.'
    )

    prepare.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_index,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively'
            ' subsets the displacement vectors. Useful if coupling is '
            'evaluated with a subset of the atoms present in the vibrational'
            ' calculation.'
        )
    )

    prepare.add_argument(
            '--red_masses',
            type=str,
            help='The reduced masses obtained from the gaussian log file using gaussian_suite extractor',
            )


    strength = subparsers.add_parser('strength')
    strength.set_defaults(func=strength_func)

    strength.add_argument(
        "cfp_file",
        type=str,
        help=(
            "File (hdf5 or CFP_polynomials) containing coupling crystal "
            "field parameters"
        )
    )

    strength.add_argument(
        "n",
        type=float,
        help=(
            "Number of unpaired electrons in 4f subshell"
        )
    )

    strength.add_argument(
        "J",
        type=float,
        help=(
            "Total angular momentum quantum number"
        )
    )

    strength.add_argument(
        "L",
        type=float,
        help=(
            "Orbital angular momentum quantum number"
        )
    )

    strength.add_argument(
        "S",
        type=float,
        help=(
            "Spin angular momentum quantum number"
        )
    )

    strength.add_argument(
        "--plot",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    strength.add_argument(
        '--irreps',
        type=str,
        metavar='<file_name>',
        help=(
            'Color code strength plot based on mode symmetries listed in file. '
            'File must contain column of IRREPs, one per mode'
        )
    )

    strength.add_argument(
        "--nooef",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    exp = subparsers.add_parser('exp')
    exp.set_defaults(func=exp_func)

    data = exp.add_argument_group('database files')

    data.add_argument(
        '-V', '--vibration_info',
        nargs='+',
        help='HDF5 database containing information about the vibrations.')

    exp.add_argument(
        '--nweights',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['coherent', 'incoherent'],
        help='Neutron scattering weights'
    )

    efield = subparsers.add_parser('efield_distort')
    efield.set_defaults(func=electric_field_func)

    state_input = efield.add_mutually_exclusive_group(required=True)

    state_input.add_argument(
        '--gaussian_fchk',
        type=str,
        help='Gaussian fchk file',
    )

    state_input.add_argument(
        '--vasp_born',
        type=str,
        help='VASP BORN file',
    )

    efield.add_argument(
        '--poscar',
        type=str,
        help='VASP POSCAR file, required for VASP BORN file',
    )

    efield.add_argument(
        '--vibrational_info',
        type=str,
        help='HDF5 database containing information about the vibrations.'
    )

    efield.add_argument(
        '--output', '-o',
        type=str,
        help='Output file name',
        required=True
    )

    field_type = efield.add_mutually_exclusive_group(required=True)

    field_type.add_argument(
        '--range',
        type=str,
        nargs=1,
        metavar=('min, max, steps, direction'),
        help='Range of field strengths, e.g. --range="start stop steps direction"'
    )

    field_type.add_argument(
        '--field_vector',
        type=float,
        nargs=3,
        metavar=('x', 'y', 'z'),
        help='Electric field vector',
    )

    field_type.add_argument(
        "--field_sampling",
        type=float,
        nargs=2,
        metavar=('Number of points', 'Field Strength'),
        help="Number of points and field strength for sampling"
    )

    # If arg_list==None, i.e. normal cli usage, parse_args() reads from
    # "sys.argv". The arg_list can be used to call the argparser from the
    # back end.

    # read sub-parser
    parser.set_defaults(func=lambda args: parser.print_help())
    _args, _ = parser.parse_known_args(arg_list)

    # select parsing option based on sub-parser
    if _args.prog in ['derivatives', 'eval', 'exp']:
        args, hpc_args = parser.parse_known_args(arg_list)
        args.func(args, hpc_args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)


def main():
    read_args()



import numpy as np
from ase.data import chemical_symbols
from hpc_suite.store import Store

from .vibrations import Harmonic


def make_exp_evaluator(vibration_info, select):

    ho = Harmonic.from_file(vibration_info)

    item, option_str = select

    options = {
        "nweights": {'approx': option_str}
    }[item]

    evaluator = {
        "nweights": NWeights
    }[item]

    return evaluator(ho, **options)


class ExpEvaluator:
    pass


class NWeights(Store):

    def __init__(self, ho, approx='incoherent'):
        self.ho = ho
        self.approx = approx

        description = ("Neutron scattering weights computed within the "
                       f"{self.approx} approximation")

        super().__init__("neutron_scattering_weights", description, label=())

    def evaluate(self):
        return compute_nweights(self.ho, approx=self.approx)

    def __iter__(self):
        yield (), self.evaluate()


# incoherent neutron scattering cross section in barn
# from Mitchell et al. (2005) Vibrational Spectroscopy with Neutrons, World Scientific
incoherent_cross_section = {
    "Ag": 0.58,
    "Al": 0.0082,
    "Ar": 0.225,
    "As": 0.06,
    "Au": 0.43,
    "B": 1.7,
    "Ba": 0.15,
    "Be": 0.0018,
    "Bi": 0.0084,
    "Br": 0.1,
    "C": 0.001,
    "Ca": 0.05,
    "Cd": 3.46,
    "Ce": 0.001,
    "Cl": 5.3,
    "Co": 4.8,
    "Cr": 1.83,
    "Cs": 0.21,
    "Cu": 0.55,
    "Dy": 54.4,
    "Er": 1.1,
    "Eu": 2.5,
    "F": 0.0008,
    "Fe": 0.4,
    "Fr": 0.0,
    "Ga": 0.16,
    "Gd": 151.0,
    "Ge": 0.18,
    "H": 80.26,
    "He": 0.0,
    "Hf": 2.6,
    "Hg": 6.6,
    "Ho": 0.36,
    "I": 0.31,
    "In": 0.54,
    "Ir": 0.0,
    "K": 0.27,
    "Kr": 0.01,
    "La": 1.13,
    "Li": 0.92,
    "Lu": 0.7,
    "Mg": 0.08,
    "Mn": 0.4,
    "Mo": 0.04,
    "N": 0.5,
    "Na": 1.62,
    "Nb": 0.0024,
    "Nd": 9.2,
    "Ne": 0.008,
    "Ni": 5.2,
    "O": 0.0008,
    "Os": 0.3,
    "P": 0.005,
    "Pa": 0.1,
    "Pb": 0.003,
    "Pd": 0.093,
    "Pm": 1.3,
    "Pr": 0.015,
    "Pt": 0.13,
    "Ra": 0.0,
    "Rb": 0.5,
    "Re": 0.9,
    "Rh": 0.3,
    "Rn": 0.0,
    "Ru": 0.4,
    "S": 0.007,
    "Sb": 0.007,
    "Sc": 4.5,
    "Se": 0.32,
    "Si": 0.004,
    "Sm": 39,
    "Sn": 0.022,
    "Sr": 0.06,
    "Ta": 0.01,
    "Tb": 0.004,
    "Tc": 0.5,
    "Te": 0.09,
    "Th": 0,
    "Ti": 2.87,
    "Tl": 0.21,
    "Tm": 0.1,
    "U": 0.005,
    "V": 5.08,
    "W": 1.63,
    "Xe": 0,
    "Y": 0.15,
    "Yb": 4,
    "Zn": 0.077,
    "Zr": 0.02
}

def compute_nweights(vib, approx='incoherent'):

    if approx == 'incoherent':
        cross_section_dict = incoherent_cross_section
    else:
        raise NotImplementedError(f"Approximation {approx} not implemented!")

    cross_secs = np.array([cross_section_dict[chemical_symbols[num]] for num in vib.atomic_nums])

    nweights = np.sum(cross_secs * np.sum(vib.displacements.real**2 + vib.displacements.imag**2, axis=2), axis=1)

    return nweights




from operator import add, mul
from functools import reduce, partial
from itertools import product
from fractions import Fraction
import numpy as np
import scipy as sp
import h5py
from ase.data import atomic_numbers
from pymatgen.core.structure import Structure
import phonopy
from phonopy.harmonic.dynmat_to_fc import get_commensurate_points
from scipy.constants import physical_constants

from vasp_suite.parser import ParseOUTCAR
import vasp_suite.hessian as outcar_hess

from gaussian_suite.extractor import make_extractor as make_gaussian_extractor

from env_suite.cells import build_cluster

import matplotlib.pyplot as plt

C0 = 299.792458e6
BOHR2M = 5.29177210903e-11
AMU = 1.66053906660e-27
HARTREE2J = 4.3597447222071e-18
EV2J = 1.602176634e-19
H_PCK = 6.62607015E-34
E0_TO_C = physical_constants['atomic unit of charge'][0]
HZ2HARTREE = 1.51983e-16


class Harmonic:
    """Set of indenpendent quantum harmonic oscillators defined by their
    frequencies, displacements and reduced masses. The coordinate system
    follows the same conventions as the Gaussian software, i.e. the cartesian
    displacements are normalised and the normalisation constant is absorbed
    into the reduced mass.

    Parameters
    ----------
    freqs : np.array
        Array of harmonic frequencies in cm^-1.
    displacements : np.array
        K x N x 3 array containing the mass-weighted displacement vectors.

    Attributes
    ----------
    freqs : np.array
        Array of harmonic frequencies in units of cm^-1.
    displacements : np.array
        K x N x 3 array containing the mass-weighted displacement vectors.
    natoms : int
        Number of atoms.
    nmodes : int
        Number of modes.
    force_const : np.array
        Array of force constants in units of mdyne/ang or N/cm.
    zpd : np.array
        Array of zero point displacements in ang.
    """

    def __init__(self, freqs, displacements, coords=None, atomic_nums=None,
                 weights=None, band_indices=None, q_points=None):

        self.nmodes = displacements.shape[0]
        self.natoms = displacements.shape[1]

        if freqs.shape[0] != self.nmodes:
            raise ValueError("Dimensions of HO parameters do not match.")

        self.freqs = freqs
        self.displacements = displacements

        self.coords = coords
        self.atomic_nums = atomic_nums

        self.weights = weights
        self.band_indices = band_indices
        self.q_points = q_points

    @property
    def mass_freq_weighted_coordinates(self):
        """Mass-frequency weighted normal mode displacements. Equivalent to the
        conversion to zero-point displacement weighted coordinates divided by
        sqrt(2)."""
        conversion = 1.0e10 / np.sqrt(AMU) * np.sqrt(H_PCK / (4 * np.pi) / self.omega)
        return self.displacements * conversion[:, np.newaxis, np.newaxis]

    @property
    def omega(self):
        """angular frequency in s^-1"""
        return 2 * np.pi * (self.freqs * 100 * C0)

    @property
    def red_masses(self):
        """Reduced masses in amu"""
        return np.sum(self.displacements**2, axis=0)

    @property
    def mode_red_masses(self):
        """Reduced masses of each mode in amu"""
        return np.sum(self.displacements**2, axis=(1,2))

    def calculate_dmudQ(self, bec_tensors):
        """
        Calculate the derivative of the polarisation with respect to
        phonon mode displacements.

        dmu/dQ = sum^n_j sum^3_b Z^*_{j,ab} X_{s,jb}

        Parameters
        ----------
        bec_tensors : ndarray of shape (N, 3, 3)
            The Born effective charge tensors.
        eigenvectors : ndarray of shape (3N, N, 3)
            The phonon eigenvectors.
        atomic_masses : ndarray of shape (N,)
            The atomic masses.

        Returns
        -------
        dmudQ : ndarray of shape (3N, 3)
            The derivative of the polarisation with respect to
            phonon mode displacements.
        """
        # tile the tensors to match the eigenvectors
        mult = int(self.displacements.shape[1] // bec_tensors.shape[0])
        if mult != 1:
            bec_tensors = np.tile(bec_tensors, (mult, 1, 1))
        assert bec_tensors.shape[0] == self.displacements.shape[1]

        # units e0 amu^{-1/2}
        dmu_dQ = np.einsum('jab, sjb -> sa', bec_tensors, self.displacements)
        # convert to C amu^{-1/2}
        dmu_dQ *= E0_TO_C
        # Normalise over the periodic images
        dmu_dQ /= mult

        return dmu_dQ

    def to_file(self, h_file, store_vecs=True):

        with h5py.File(h_file, 'w') as h:
            h['/'].attrs.create('num_modes', self.nmodes)
            h['/'].attrs.create('num_atoms', self.natoms)
            h.create_dataset('frequencies', data=self.freqs)

            if store_vecs:
                h.create_dataset('displacements', data=self.displacements)

            if self.coords is not None:
                h.create_dataset('coords', data=self.coords)

            if self.atomic_nums is not None:
                h.create_dataset('atomic_nums', data=self.atomic_nums)

            if self.weights is not None:
                h.create_dataset('weights', data=self.weights)

            if self.band_indices is not None:
                h.create_dataset('band_indices', data=self.band_indices)

            if self.q_points is not None:
                h.create_dataset('q_points', data=self.q_points)

    def to_pyvibms(self, file):
        with open(file, 'w') as f:
            f.write(f"{self.natoms} {self.nmodes}\n")
            for idx, (freq, displacement) in enumerate(zip(self.freqs, self.displacements), start=1):
                normalised = displacement.real / np.linalg.norm(displacement)
                f.write(f"\nN {freq:9.4f} A {idx}\n")
                f.write('\n'.join(f"{displ:9.4f}" for vec in normalised for displ in vec))
                f.write('\n')
            f.write("END\n")

    def subset_atoms(self, active_atom_idc=None, ref_coords=None):

        def generate_mapping():
            for idx, coord in enumerate(ref_coords):
                idc = np.flatnonzero(np.isclose(self.coords, coord).all(axis=1))
                if len(idc) == 0:
                    continue
                elif len(idc) == 1:
                    yield idc[0]
                else:
                    raise ValueError(f"Multiple instances of atom {coord}!")

        if active_atom_idc is not None and ref_coords is not None:
            raise ValueError("Supply one of active_atom_idc or ref_coords!")
        elif active_atom_idc is not None:
            idc = active_atom_idc
        elif ref_coords is not None:
            idc = list(generate_mapping())
        else:
            return self

        coords = None if self.coords is None else self.coords[idc]
        atomic_nums = None if self.atomic_nums is None else self.atomic_nums[idc]

        return self.__class__(self.freqs, self.displacements[:, idc],
                              coords=coords, atomic_nums=atomic_nums,
                              weights=self.weights, band_indices=self.band_indices, q_points=self.q_points)

    @classmethod
    def from_file(cls, h_file):

        with h5py.File(h_file, 'r') as h:
            displacements = h['displacements'][...]
            freqs = h['frequencies'][...]

            try:
                coords = h['coords'][...]
            except KeyError:
                coords = None

            try:
                atomic_nums = h['atomic_nums'][...]
            except KeyError:
                atomic_nums = None

            try:
                weights = h['weights'][...]
            except KeyError:
                weights = None

            try:
                band_indices = h['band_indices'][...]
            except KeyError:
                band_indices = None

            try:
                q_points = h['q_points'][...]
            except KeyError:
                q_points = None

        return cls(freqs, displacements,
                   coords=coords, atomic_nums=atomic_nums,
                   weights=weights, band_indices=band_indices, q_points=q_points)

    @classmethod
    def from_gaussian_log(cls, f):
        freqs = make_gaussian_extractor(f, ("freq", "frequency"))[1]
        modes = make_gaussian_extractor(f, ("freq", "displacement"))[1]
        red_masses = make_gaussian_extractor(f, ("freq", "reduced_mass"))[1]
        return cls(freqs, modes / np.sqrt(red_masses)[:, np.newaxis, np.newaxis])

    @classmethod
    def from_gaussian_fchk(cls, f, ext_masses=None, **kwargs):
        hess = make_gaussian_extractor(f, ('fchk', 'hessian'))[()]
        masses = make_gaussian_extractor(f, ('fchk', 'atomic_mass'))[()]
        numbers = make_gaussian_extractor(f, ('fchk', 'atomic_number'))[()]
        coords = make_gaussian_extractor(f, ('fchk', 'coordinates'))[()]

        for key, mass in ext_masses.items():
            if isinstance(key, int):
                masses[key - 1] = mass
            elif isinstance(key, str):
                num = atomic_numbers[key]
                for idx in [i for i, n in enumerate(numbers) if n == num]:
                    masses[idx] = mass
            else:
                raise KeyError("Specify isotopic substitutions with either "
                               "element symbol (str) or atomic indices (int)!")

        return cls.analysis(hess, masses, coords=coords * BOHR2M * 1e10,
                            atomic_nums=numbers, **kwargs)

    @classmethod
    def from_orca_hess(cls, f, ext_masses=None, **kwargs):
        from orca_suite.extractor import make_extractor as make_orca_extractor
        hess = make_orca_extractor(f, ('hess', 'hessian'))[()]
        masses = make_orca_extractor(f, ('hess', 'atomic_mass'))[()]
        symbols = make_orca_extractor(f, ('hess', 'atomic_symbol'))[()]
        numbers = [atomic_numbers[sym.decode()] for sym in symbols]
        coords = make_orca_extractor(f, ('hess', 'coordinates'))[()]

        for key, mass in ext_masses.items():
            if isinstance(key, int):
                masses[key - 1] = mass
            elif isinstance(key, str):
                num = atomic_numbers[key]
                for idx in [i for i, n in enumerate(numbers) if n == num]:
                    masses[idx] = mass
            else:
                raise KeyError("Specify isotopic substitutions with either "
                               "element symbol (str) or atomic indices (int)!")

        return cls.analysis(hess, masses, coords=coords * BOHR2M * 1e10,
                            atomic_nums=numbers, **kwargs)

    @classmethod
    def from_vasp_outcar(cls, outcar_file, poscar_file, trans=True, **kwargs):
        outcar = ParseOUTCAR(outcar_file, poscar_file=poscar_file)
        freqs = outcar_hess.parse_frequencies(outcar)
        displacements = outcar_hess.parse_displacements(outcar)
        freqs, displacements = zip(*sorted(zip(freqs, displacements), key=lambda x: x[0]))
        freqs = np.array(freqs)
        displacements = np.array(displacements)

        if trans:
            freqs = freqs[3:]
            displacements = displacements[3:]

        masses = outcar.masses
        coords = outcar.cart_coords
        return cls(freqs, displacements / np.sqrt(masses)[np.newaxis, :, np.newaxis], coords=coords, **kwargs)

    @classmethod
    def generate_from_qpoints(cls, q_mesh, q_points, freqs_list, vecs_list,
                              poscar, masses, trans=True, **cluster_kwargs):

        cluster = build_cluster(poscar, **cluster_kwargs)

        n_cell = cluster.n_cell

        cart = cluster.cart_coords
        frac = cluster.frac_coords

        # norm is sqrt(#qpoint)
        norm = np.sqrt(reduce(mul, q_mesh))

        for q_point, freqs, vecs in zip(q_points, freqs_list, vecs_list):

            nmodes, natoms = vecs.shape[1], vecs.shape[0] // 3

            # reshape to K x N x 3 (K: #modes, N: #atoms)
            unitcell_displacements = vecs.T.reshape(nmodes, natoms, 3) / \
                np.sqrt(masses)[np.newaxis, :, np.newaxis] / norm

            if trans and q_point.is_gamma():
                freqs = freqs[3:]
                unitcell_displacements = unitcell_displacements[3:]

            print(f"Evaluating q-point {q_point}")
            
            # With boundary atoms, need to map cluster atoms to their POSCAR index
            # using "ENV" labels
            try:
                poscar_index = np.array([int(label.split('.')[-1]) - 1 for label in cluster.labels])
            except (ValueError, IndexError) as e: 
                raise ValueError(f"Could not parse original POSCAR index from cluster labels. Error: {e}")

            # Map unit cell displacements, atomic numbers to the cluster:
            displacements = unitcell_displacements[:, poscar_index, :]
            atomic_nums = np.array(cluster.atom_numbers)[poscar_index]

            # apply relative phases at off-gamma q-points
            if not q_point.is_gamma():
                phase = np.exp(2.j * np.pi * frac @ q_point.coords)
                displacements *= phase[np.newaxis, :, np.newaxis]

            nmodes = freqs.size

            if q_point.is_boundary() or q_point.is_gamma():

                abs_phase = np.array(list(map(phase_by_min_imag, displacements)))
                displacements *= abs_phase[:, np.newaxis, np.newaxis]

                imag_norm = np.linalg.norm(displacements.imag)
                real_norm = np.linalg.norm(displacements.real)
                print(f"{q_point}-point displacements have non-zero "
                      f"imaginary component: {imag_norm}, "
                      f"real component: {real_norm}!")

                yield cls(freqs, displacements.real,
                          coords=cart, atomic_nums=atomic_nums,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=np.tile(q_point.coords, (nmodes, 1)))
            else:

                yield cls(freqs, np.sqrt(2) * displacements.real,
                          coords=cart, atomic_nums=atomic_nums,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=np.tile(q_point.coords, (nmodes, 1)))

                yield cls(freqs, np.sqrt(2) * displacements.imag,
                          coords=cart, atomic_nums=atomic_nums,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=-np.tile((-q_point).coords, (nmodes, 1)))

    @classmethod
    def from_vasp_phonopy(cls, poscar, force_sets, ext_masses=None, trans=True,
                          force_expansion=(1, 1, 1), q_mesh=(1, 1, 1),
                          **cluster_kwargs):
        """Evaluate harmonic oscillators from VASP-phonopy calculation based
        on 
        """

        q_points = list(reduce_half_bz(map(partial(Qpoint, q_mesh), product(*map(range, q_mesh)))))

        cell = phonopy.load(
            unitcell_filename=poscar,
            force_sets_filename=force_sets,
            supercell_matrix=force_expansion,
            primitive_matrix=np.identity(3)
        )

        masses = cell._dynamical_matrix._pcell.masses
        symbols = cell.unitcell.symbols

        for key, mass in ext_masses.items():
            if isinstance(key, int):
                masses[key - 1] = mass
            elif isinstance(key, str):
                for idx in [i for i, sym in enumerate(symbols) if sym == key]:
                    masses[idx] = mass
            else:
                raise KeyError("Specify isotopic substitutions with either "
                               "element symbol (str) or atomic indices (int)!")

        if ext_masses:
            cell._dynamical_matrix._pcell.masses = masses
            cell._set_dynamical_matrix()

        cell.run_qpoints([q.coords for q in q_points], with_eigenvectors=True)

        qpoint_dict = cell.get_qpoints_dict()

        modes = Harmonic.generate_from_qpoints(
            q_mesh,
            q_points,
            qpoint_dict['frequencies'] * 1e12 / (C0 * 1e2),
            qpoint_dict['eigenvectors'],
            poscar,
            masses,
            trans=trans,
            **cluster_kwargs)

        return reduce(add, modes)

    @classmethod
    def from_phonopy_hdf5(cls, poscar, h_file, trans=True, **cluster_kwargs):

        with h5py.File(h_file, 'r') as h:

            q_mesh = tuple(h['mesh'][...])
            q_point_list_full = list(map(partial(Qpoint, q_mesh),
                                         map(tuple, h['grid_address'][...])))
            freqs_list_full = h['frequency'][...] * 1e12 / (C0 * 1e2)
            vecs_list_full = h['eigenvector'][...]

        unique_half_bz = list(map(lambda q: q.address, reduce_half_bz(q_point_list_full)))

        q_point_list, freqs_list, vecs_list = zip(*filter(
            lambda args: args[0].address in unique_half_bz,
            zip(q_point_list_full, freqs_list_full, vecs_list_full)))

        structure = Structure.from_file(poscar)
        masses = np.array([site.specie.atomic_mass for site in structure.sites])

        modes = Harmonic.generate_from_qpoints(
            q_mesh,
            q_point_list,
            freqs_list,
            vecs_list,
            poscar,
            masses,
            trans=trans,
            **cluster_kwargs
        )

        return reduce(add, modes)

    @property
    def hessian(self):
        """
        Reform the Hessian matrix

        self.omega = Omega^2
        Omega^2 = L^T D^T sqrt{M^-1} H sqrt{M^-1} D L

        H = sqrt{M} D L Omega^2 L^T D^T sqrt{M}

        where displacements = sqrt{M^{-1}} D L

        parameters
        ----------
        self : Harmonic

        returns
        -------
        H : np.ndarray
        """
        disp_shape = self.displacements.shape
        displacements = self.displacements.reshape(disp_shape[0], -1)
        # displacements units: sqrt(amu)
        omega = self.omega  # rad s^-1
        omega2 = omega ** 2  # rad^2 s^-2

        hessian = displacements.T @ np.diag(omega2) @ displacements
        # units = sqrt(amu) * rad^2 s^-2 * sqrt(amu)
        # units = amu s^-2
        hessian *= AMU  # Kg s^-2 == J m^-2
        # hessian *= (1/HARTREE2J) * (1/BOHR2M) ** 2  # Ha Bohr^-2
        return hessian

    def remove_imaginary(self, verbose=False):

        def is_real(idx_freq):
            idx, freq = idx_freq
            if verbose and freq < 0:
                print(f"Removing imaginary mode: {idx+1} ({freq: .2f} cm^-1)")
                return False
            else:
                return True

        idc = list(tuple(zip(*filter(is_real, enumerate(self.freqs))))[0])

        return Harmonic(self.freqs[idc], self.displacements[idc],
                        coords=self.coords, atomic_nums=self.atomic_nums,
                        weights=None if self.weights is None else self.weights[idc],
                        band_indices=None if self.band_indices is None else self.band_indices[idc],
                        q_points=None if self.q_points is None else self.q_points[idc])

    def abs_imaginary(self, verbose=False):
        
        if verbose: 
            for idx, freq in enumerate(self.freqs):
                if freq < 0: print(f"Setting imaginary mode: {idx+1} {freq: .2f} cm^-1 to {abs(freq): .2f} cm^-1")

        self.freqs=abs(self.freqs)
        idc=np.argsort(self.freqs)

        return Harmonic(self.freqs[idc], self.displacements[idc],
                        coords=self.coords, atomic_nums=self.atomic_nums,
                        weights=None if self.weights is None else self.weights[idc],
                        band_indices=None if self.band_indices is None else self.band_indices[idc],
                        q_points=None if self.q_points is None else self.q_points[idc])

    def __add__(self, other):

        if not np.allclose(self.coords, other.coords):
            raise ValueError("Coordinates are incompatible!")

        if not all(self.atomic_nums == other.atomic_nums):
            raise ValueError("Atomic numbers are incompatible!")

        order = np.argsort(np.concatenate((self.freqs, other.freqs)))

        def concat_order(attr):

            self_attr = getattr(self, attr)
            other_attr = getattr(other, attr)

            if self_attr is None or other_attr is None:
                return None
            else:
                return np.concatenate((self_attr, other_attr))[order]

        freqs = concat_order("freqs")
        displacements = concat_order("displacements")

        coords = self.coords
        atomic_nums = self.atomic_nums

        weights = concat_order("weights")
        band_indices = concat_order("band_indices")
        q_points = concat_order("q_points")

        return self.__class__(freqs, displacements, coords=coords, atomic_nums=atomic_nums,
                              weights=weights, band_indices=band_indices,
                              q_points=q_points)

    @classmethod
    def analysis(cls, hessian, masses, trans=True, rot=True,
                 coords=None, atomic_nums=None):

        natom = len(masses)

        mwhess = hessian / np.sqrt(np.repeat(masses, 3)[np.newaxis, :] *
                                   np.repeat(masses, 3)[:, np.newaxis])

        eig, vec = np.linalg.eigh(mwhess)

        if trans:
            tra_frame = [np.array([d * np.sqrt(m) for m in masses for d in v])
                         for v in np.identity(3)]
            ntra = 3
        else:
            tra_frame = []
            ntra = 0

        if rot:
            # compute principle axis frame
            nrot, _, vec_inertia = principle_axis_inertia(masses, coords)

            # convert coordinates to principle axis frame
            # _coords = shift_to_com(masses, coords) @ vec_inertia.T
            _coords = shift_to_com(masses, coords)

            rot_frame = \
                [np.array([d * np.sqrt(m) for m, c in zip(masses, _coords)
                           for d in np.cross(c, v)])
                 for v in vec_inertia[:, (3 - nrot):].T]
        else:
            rot_frame = []
            nrot = 0

        nrotra = ntra + nrot
        nmodes = 3 * natom - nrotra

        rotra_frame = \
            np.array([d / np.linalg.norm(d) for d in tra_frame + rot_frame])

        if len(rotra_frame) != 0:
            # detect rigid body motions
            int_mask = np.logical_not(np.isclose(
                np.sum((rotra_frame @ vec)**2, axis=0), 1.0, rtol=1e-1))
            # Schmidt orthogonalisation
            new_frame, _ = np.linalg.qr(
                np.column_stack((rotra_frame.T, vec[:, int_mask])))
            # set coordinate frame to internal coordiantes
            frame = new_frame[:, nrotra:]

            # transfrom Hessian to internal coordinate frame and project out
            # rigid body motions
            eig, vec = np.linalg.eigh(frame.T @ mwhess @ frame)
            cart = frame @ vec / np.sqrt(np.repeat(masses, 3)[:, np.newaxis])

        else:
            cart = vec / np.sqrt(np.repeat(masses, 3)[:, np.newaxis])

        freqs = np.sqrt(eig * (HARTREE2J / BOHR2M**2 / AMU) /
                        (4*np.pi**2 * C0**2) + 0.j) / 100
        _freqs = np.vectorize(lambda x: -x.imag if x.imag else x.real)(freqs)

        displacements = np.reshape(cart.T, (nmodes, natom, 3))

        return cls(_freqs, displacements,
                   coords=coords, atomic_nums=atomic_nums)


class Qpoint:

    def __init__(self, mesh, address):
        self.mesh = mesh
        self.address = address

    @property
    def coords(self):
        return np.array(list(map(lambda n, x: x / n, self.mesh, self.address)))

    def __eq__(self, other):

        if self.mesh != other.mesh:
            ValueError(f"Comparing unequal q-meshes: {self.mesh} {other.mesh}")

        def predicate(n, x, y):  # same or translationally equivalent
            return (x - y) % n == 0

        return all(map(predicate, self.mesh, self.address, other.address))

    def is_gamma(self):

        def predicate(n, x):
            return x % n == 0
        
        return all(map(predicate, self.mesh, self.address))

    def is_boundary(self):
        
        def predicate(n, x):
            return x % n == 0 or (n % 2 == 0 and x % n == n // 2)
        
        return all(map(predicate, self.mesh, self.address)) and not self.is_gamma()

    def __ne__(self, other):
        return not self == other

    def __neg__(self):
        return Qpoint(self.mesh, tuple(map(lambda x: -x, self.address)))

    def __hash__(self):
        return hash(tuple(x % n for n, x in zip(self.mesh, self.address)))

    def __str__(self):
        return str(self.coords)

    def __repr__(self):
        return f"Qpoint({self.mesh},{self.address})"


def reduce_half_bz(q_points):

    unique_q_points = set()

    for q_point in q_points:

        if q_point in unique_q_points or -q_point in unique_q_points:
            continue
        else:
            unique_q_points.add(q_point)
            yield q_point


def phase_by_min_imag(x):

    ang = np.angle(x)
    mag = np.abs(x)**2

    phase = -np.arctan(np.sum(mag * np.sin(2 * ang)) / np.sum(mag * np.cos(2 * ang))) / 2

    if 2 * np.sum(mag * np.cos(2 * (ang + phase))) > 0:
        return np.exp(1.j * phase)
    else:
        return np.exp(1.j * (phase + np.pi / 2))


def shift_to_com(masses, coords):
    com = np.sum(masses[:, np.newaxis] * coords, axis=0) / np.sum(masses)
    return coords - com


def principle_axis_inertia(masses, coords):

    _coords = shift_to_com(masses, coords)

    inertia = np.zeros((3, 3))
    inertia[0, 0] = np.sum(masses * (_coords[:, 1]**2 + _coords[:, 2]**2))
    inertia[1, 1] = np.sum(masses * (_coords[:, 2]**2 + _coords[:, 0]**2))
    inertia[2, 2] = np.sum(masses * (_coords[:, 0]**2 + _coords[:, 1]**2))
    inertia[1, 0] = -np.sum(masses * (_coords[:, 0] * _coords[:, 1]))
    inertia[2, 0] = -np.sum(masses * (_coords[:, 0] * _coords[:, 2]))
    inertia[2, 1] = -np.sum(masses * (_coords[:, 1] * _coords[:, 2]))
    inertia[0, 1] = inertia[1, 0]
    inertia[0, 2] = inertia[2, 0]
    inertia[1, 2] = inertia[2, 1]

    eig_inertia, vec_inertia = np.linalg.eig(inertia)
    rank = np.linalg.matrix_rank(inertia)

    if rank > 1:
        return rank, eig_inertia, vec_inertia
    else:
        raise ValueError("Rank of moment of inertia tensor smaller than one.")


def construct_atomic_basis(dim, gaussian_fchk=False, vasp_born=False, coords=None):
    def make_rigid_translations():
        x = np.array([np.eye(3)[0]] * dim).reshape(-1, 1)
        y = np.array([np.eye(3)[1]] * dim).reshape(-1, 1)
        z = np.array([np.eye(3)[2]] * dim).reshape(-1, 1)
        return np.hstack((x, y, z))

    def make_rigid_rotations(coords):
        x = np.array(list(map(lambda x: np.cross([1, 0, 0], x), coords))).reshape(-1, 1)
        y = np.array(list(map(lambda x: np.cross([0, 1, 0], x), coords))).reshape(-1, 1)
        z = np.array(list(map(lambda x: np.cross([0, 0, 1], x), coords))).reshape(-1, 1)
        return np.hstack((x, y, z))

    rigid_translations = make_rigid_translations()
    if gaussian_fchk:
        if coords is None:
            raise ValueError('Supply coordinates for rigid body rotations')
        rigid_rotations = make_rigid_rotations(coords)
        energy_invariant = np.hstack((rigid_translations, rigid_rotations))
    else:
        energy_invariant = rigid_translations

    q, _, _ = sp.linalg.qr(energy_invariant, mode='full', pivoting=True)
    if gaussian_fchk:
        return q[6:].T
    else:
        return q[3:].T



"""
A module to calculate the derivative of the
polarisation with respect to phonon mode displacements.
"""
import numpy as np

from phonopy.file_IO import parse_BORN
from phonopy.structure.atoms import Atoms
from vasp_suite.structure import Structure

import jax.numpy as jnp
from jax import grad, jit

#
#           Notes:
#           ------
#           Dependency on spglib version 2.0.2,
#           Newer versions of spglib are not compatible with phonopy
#
#           dmu_dQ is in units of [C].
#           This makes line 2303 of firm_sim.f90 redundant
#           Conditional if statement required to not convert units
#
#


def parse_BEC(poscar_file, born_file):
    """
    Parse the Born effective charge tensors from the 'BORN' file
    produced by Phonopy.

    Parameters
    ----------
    poscar_file : str
        The path to the POSCAR file.

    Returns
    -------
    BEC : ndarray of shape (N, 3, 3)
    """
    poscar = Structure.from_poscar(poscar_file)
    cell = Atoms(symbols=poscar.atom_list, cell=poscar.lattice_vectors,
                 scaled_positions=poscar.coords)
    born = parse_BORN(cell, filename=born_file)
    return np.array([tensor for tensor in born['born']])


def normalised_sphere_vectors(num_points):
    polar_vectors = sample_sphere(num_points)
    cartesian_vectors = np.array(list(map(polar_to_cartesian, polar_vectors)))
    return cartesian_vectors


def polar_to_cartesian(polar_vector):
    theta, phi = polar_vector
    x = 1 * np.sin(theta) * np.cos(phi)
    y = 1 * np.sin(theta) * np.sin(phi)
    z = 1 * np.cos(theta)
    return np.array([x, y, z])


def sample_sphere(num_points):
    angles = np.linspace(0, 2 * np.pi, num_points)
    polar_vectors = list(map(lambda x: list(map(lambda y: [x,  y], angles)), angles))
    return np.array(polar_vectors).reshape(-1, 2)



import numpy as np

def print_tau_style(derivatives, freqs, f_name):

    with open(f_name, 'w') as f:
        f.truncate(0)

        for freq, ((idx, _, _), pars) in zip(freqs, derivatives):

            f.write(("Vibration: {:d} Energy (cm-1): {} Rank (k), order (q) "
                     "and derivatives up to 3rd order\n").format(idx, freq))

            for (k, q), Bkq_dq in pars.items():
                f.write(f"{k:^ d} {q:^ d}  {0: 14.9f}  {0: 14.9f}  {Bkq_dq: 14.9f}\n")


def read_tau_style(f_name: str):
    """
    Reads CFP_polynomials.dat and extracts coupling values and mode energies

    Parameters
    ----------
    f_name : str
        CFP_polynomials file name

    Returns
    -------
    np.ndarray
        Polynomial coefficients a, b, c for each mode (3, n_modes)
    list
        Mode energies
    """

    freqs = []
    derivatives = []

    with open(f_name, 'r') as f:

        for line in f:
            if "Vibration:" in line:
                freqs.append(float(line.split()[4]))
                tmp = []
                for _ in range(27):
                    line = next(f)
                    tmp.append(float(line.split()[-1]))
                derivatives.append(tmp)

    freqs = np.array(freqs)
    derivatives = np.array(derivatives)

    return derivatives, freqs



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long,missing-module-docstring,exec-used

import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

# DO NOT EDIT THIS NUMBER!
# IT IS AUTOMATICALLY CHANGED BY python-semantic-release
__version__ = "1.14.1"

setuptools.setup(
    name='spin_phonon_suite',
    version=__version__,
    author='Chilton Group',
    author_email='nicholas.chilton@manchester.ac.uk',
    description='A package for performing spin-phonon coupling calculations with openMOLCAS, VASP, and Gaussian', # noqa
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://gitlab.com/chilton-group/spin_phonon_suite",
    project_urls={
        "Bug Tracker": "https://gitlab.com/chilton-group/spin_phonon_suite/-/issues", # noqa
        "Documentation": "https://chilton-group.gitlab.io/spin_phonon_suite"
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent'
        ],
    python_requires='>=3.9',
    install_requires=[
        'numpy < 2.0.0',
        'scipy',
        'h5py',
        'xyz_py>=5.1.0',
        'angmom_suite>=1.22.0',
        'hpc_suite>=1.8.0',
        'matplotlib',
        'gaussian_suite>=1.11.0',
        'phonopy==2.22.0',
        'molvis>=0.3.0',
        'molcas_suite>=1.21.0',
        'ase',
        'env_suite',
        'vasp_suite>=1.9.0',
        'spglib==2.0.2',
        ],
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'spin_phonon_suite = spin_phonon_suite.cli:main'
            ]
        }
    )
